@@ -10,50 +10,25 @@ import py
 import pytest
 
 import tox
-from tox._pytestplugin import ReportExpectMock
 from tox.config import parseconfig
+from tox.reporter import Verbosity
 from tox.session import Session
 
 pytest_plugins = "pytester"
 
 
-def test_report_protocol(newconfig):
-    config = newconfig(
-        [],
-        """
-            [testenv:mypython]
-            deps=xy
-    """,
-    )
-
-    class Popen:
-        def __init__(self, *args, **kwargs):
-            pass
-
-        def communicate(self):
-            return "", ""
-
-        def wait(self):
-            pass
-
-    session = Session(config, popen=Popen, Report=ReportExpectMock)
-    report = session.report
-    report.expect("using")
-    venv = session.getvenv("mypython")
-    action = session.newaction(venv, "update")
-    venv.update(action)
-    report.expect("logpopen")
-
-
 class TestSession:
     def test_log_pcall(self, mocksession):
+        mocksession.logging_levels(quiet=Verbosity.DEFAULT, verbose=Verbosity.INFO)
         mocksession.config.logdir.ensure(dir=1)
         assert not mocksession.config.logdir.listdir()
-        action = mocksession.newaction(None, "something")
-        action.popen(["echo"])
-        match = mocksession.report.getnext("logpopen")
-        assert match[1].outpath.relto(mocksession.config.logdir)
-        assert match[1].shell is False
+        with mocksession.newaction("what", "something") as action:
+            action.popen(["echo"])
+            match = mocksession.report.getnext("logpopen")
+            log_name = py.path.local(match[1].split(">")[-1].strip()).relto(
+                mocksession.config.logdir
+            )
+            assert log_name == "what-0.log"
 
     def test_summary_status(self, initproj, capfd):
         initproj(
@@ -68,7 +43,7 @@ class TestSession:
         )
         config = parseconfig([])
         session = Session(config)
-        envs = session.venvlist
+        envs = list(session.venv_dict.values())
         assert len(envs) == 2
         env1, env2 = envs
         env1.status = "FAIL XYZ"
@@ -112,14 +87,14 @@ def test_notoxini_help_still_works(initproj, cmd):
     assert result.err == msg
     assert result.out.startswith("usage: ")
     assert any("--help" in l for l in result.outlines), result.outlines
-    assert not result.ret
+    result.assert_success(is_run_test_env=False)
 
 
 def test_notoxini_help_ini_still_works(initproj, cmd):
     initproj("example123-0.5", filedefs={"tests": {"test_hello.py": "def test_hello(): pass"}})
     result = cmd("--help-ini")
     assert any("setenv" in l for l in result.outlines), result.outlines
-    assert not result.ret
+    result.assert_success(is_run_test_env=False)
 
 
 def test_envdir_equals_toxini_errors_out(cmd, initproj):
@@ -134,8 +109,10 @@ def test_envdir_equals_toxini_errors_out(cmd, initproj):
     )
     result = cmd()
     assert result.outlines[1] == "ERROR: ConfigError: envdir must not equal toxinidir"
-    assert re.match(r"ERROR: venv \'python\' in .* would delete project", result.outlines[0])
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    assert re.match(
+        r"ERROR: venv \'python\' in .* would delete project", result.outlines[0]
+    ), result.outlines[0]
+    result.assert_fail()
 
 
 def test_run_custom_install_command_error(cmd, initproj):
@@ -149,10 +126,11 @@ def test_run_custom_install_command_error(cmd, initproj):
         },
     )
     result = cmd()
-    assert re.match(
-        r"ERROR: invocation failed \(errno \d+\), args: .*[/\\]tox\.ini", result.outlines[-1]
-    )
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
+    re.match(
+        r"ERROR:   python: InvocationError for command .* \(exited with code \d+\)",
+        result.outlines[-1],
+    ), result.out
 
 
 def test_unknown_interpreter_and_env(cmd, initproj):
@@ -165,17 +143,18 @@ def test_unknown_interpreter_and_env(cmd, initproj):
             basepython=xyz_unknown_interpreter
             [testenv]
             changedir=tests
+            skip_install = true
         """,
         },
     )
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert any(
         "ERROR: InterpreterNotFound: xyz_unknown_interpreter" == l for l in result.outlines
     ), result.outlines
 
     result = cmd("-exyz")
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert result.out == "ERROR: unknown environment 'xyz'\n"
 
 
@@ -193,7 +172,7 @@ def test_unknown_interpreter(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert any(
         "ERROR: InterpreterNotFound: xyz_unknown_interpreter" == l for l in result.outlines
     ), result.outlines
@@ -212,9 +191,11 @@ def test_skip_platform_mismatch(cmd, initproj):
         },
     )
     result = cmd()
-    assert not result.ret
+    result.assert_success()
     assert any(
-        "SKIPPED:  python: platform mismatch" == l for l in result.outlines
+        "SKIPPED:  python: platform mismatch ({!r} does not match 'x123')".format(sys.platform)
+        == l
+        for l in result.outlines
     ), result.outlines
 
 
@@ -232,7 +213,7 @@ def test_skip_unknown_interpreter(cmd, initproj):
         },
     )
     result = cmd("--skip-missing-interpreters")
-    assert not result.ret
+    result.assert_success()
     msg = "SKIPPED:  python: InterpreterNotFound: xyz_unknown_interpreter"
     assert any(msg == l for l in result.outlines), result.outlines
 
@@ -252,13 +233,13 @@ def test_skip_unknown_interpreter_result_json(cmd, initproj, tmpdir):
         },
     )
     result = cmd("--skip-missing-interpreters", "--result-json", report_path)
-    assert not result.ret
+    result.assert_success()
     msg = "SKIPPED:  python: InterpreterNotFound: xyz_unknown_interpreter"
     assert any(msg == l for l in result.outlines), result.outlines
     setup_result_from_json = json.load(report_path)["testenvs"]["python"]["setup"]
     for setup_step in setup_result_from_json:
         assert "InterpreterNotFound" in setup_step["output"]
-        assert setup_step["retcode"] == "0"
+        assert setup_step["retcode"] == 0
 
 
 def test_unknown_dep(cmd, initproj):
@@ -274,7 +255,7 @@ def test_unknown_dep(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert result.outlines[-1].startswith("ERROR:   python: could not install deps [qweqwe123];")
 
 
@@ -292,7 +273,7 @@ def test_venv_special_chars_issue252(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret == 0, "{}\n{}".format(result.err, result.out)
+    result.assert_success()
     pattern = re.compile("special&&1 installed: .*pkg123==0.7.*")
     assert any(pattern.match(line) for line in result.outlines), result.outlines
 
@@ -300,10 +281,25 @@ def test_venv_special_chars_issue252(cmd, initproj):
 def test_unknown_environment(cmd, initproj):
     initproj("env123-0.7", filedefs={"tox.ini": ""})
     result = cmd("-e", "qpwoei")
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert result.out == "ERROR: unknown environment 'qpwoei'\n"
 
 
+def test_unknown_environment_with_envlist(cmd, initproj):
+    initproj(
+        "pkg123",
+        filedefs={
+            "tox.ini": """
+            [tox]
+            envlist = py{36,37}-django{20,21}
+        """
+        },
+    )
+    result = cmd("-e", "py36-djagno21")
+    result.assert_fail()
+    assert result.out == "ERROR: unknown environment 'py36-djagno21'\n"
+
+
 def test_minimal_setup_py_empty(cmd, initproj):
     initproj(
         "pkg123-0.7",
@@ -315,7 +311,7 @@ def test_minimal_setup_py_empty(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret == 1, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert result.outlines[-1] == "ERROR: setup.py is empty"
 
 
@@ -331,7 +327,7 @@ def test_minimal_setup_py_comment_only(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret == 1, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert result.outlines[-1] == "ERROR: setup.py is empty"
 
 
@@ -348,7 +344,7 @@ def test_minimal_setup_py_non_functional(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret == 1, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert any(re.match(r".*ERROR.*check setup.py.*", l) for l in result.outlines), result.outlines
 
 
@@ -364,7 +360,7 @@ def test_sdist_fails(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert any(
         re.match(r".*FAIL.*could not package project.*", l) for l in result.outlines
     ), result.outlines
@@ -382,7 +378,7 @@ def test_no_setup_py_exits(cmd, initproj):
     )
     os.remove("setup.py")
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert any(
         re.match(r".*ERROR.*No setup.py file found.*", l) for l in result.outlines
     ), result.outlines
@@ -409,7 +405,7 @@ def test_package_install_fails(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert result.outlines[-1].startswith("ERROR:   python: InvocationError for command ")
 
 
@@ -437,12 +433,12 @@ def example123(initproj):
 
 def test_toxuone_env(cmd, example123):
     result = cmd()
-    assert not result.ret
+    result.assert_success()
     assert re.match(
         r".*generated\W+xml\W+file.*junit-python\.xml" r".*\W+1\W+passed.*", result.out, re.DOTALL
     )
     result = cmd("-epython")
-    assert not result.ret
+    result.assert_success()
     assert re.match(
         r".*\W+1\W+passed.*" r"summary.*" r"python:\W+commands\W+succeeded.*",
         result.out,
@@ -450,11 +446,11 @@ def test_toxuone_env(cmd, example123):
     )
 
 
-def test_different_config_cwd(cmd, example123, monkeypatch):
+def test_different_config_cwd(cmd, example123):
     # see that things work with a different CWD
-    monkeypatch.chdir(example123.dirname)
-    result = cmd("-c", "example123/tox.ini")
-    assert not result.ret
+    with example123.dirpath().as_cwd():
+        result = cmd("-c", "example123/tox.ini")
+    result.assert_success()
     assert re.match(
         r".*\W+1\W+passed.*" r"summary.*" r"python:\W+commands\W+succeeded.*",
         result.out,
@@ -462,19 +458,49 @@ def test_different_config_cwd(cmd, example123, monkeypatch):
     )
 
 
-def test_json(cmd, example123):
-    # see that tests can also fail and retcode is correct
-    testfile = py.path.local("tests").join("test_hello.py")
-    assert testfile.check()
-    testfile.write("def test_fail(): assert 0")
-    jsonpath = example123.join("res.json")
-    result = cmd("--result-json", jsonpath)
-    assert result.ret == 1, "{}\n{}".format(result.err, result.out)
-    data = json.load(jsonpath.open("r"))
-    verify_json_report_format(data)
-    assert re.match(
-        r".*\W+1\W+failed.*" r"summary.*" r"python:\W+commands\W+failed.*", result.out, re.DOTALL
+def test_result_json(cmd, initproj, example123):
+    cwd = initproj(
+        "example123",
+        filedefs={
+            "tox.ini": """
+            [testenv]
+            deps = setuptools
+            commands_pre = python -c 'print("START")'
+            commands = python -c 'print("OK")'
+                       - python -c 'print("1"); raise SystemExit(1)'
+                       python -c 'print("1"); raise SystemExit(2)'
+                       python -c 'print("SHOULD NOT HAPPEN")'
+            commands_post = python -c 'print("END")'
+        """
+        },
     )
+    json_path = cwd / "res.json"
+    result = cmd("--result-json", json_path)
+    result.assert_fail()
+    data = json.loads(json_path.read_text(encoding="utf-8"))
+
+    assert data["reportversion"] == "1"
+    assert data["toxversion"] == tox.__version__
+
+    for env_data in data["testenvs"].values():
+        for command_type in ("setup", "test"):
+            if command_type not in env_data:
+                assert False, "missing {}".format(command_type)
+            for command in env_data[command_type]:
+                assert isinstance(command["command"], list)
+                assert command["output"]
+                assert "retcode" in command
+                assert isinstance(command["retcode"], int)
+        # virtualenv, deps install, package install, freeze
+        assert len(env_data["setup"]) == 4
+        # 1 pre + 3 command + 1 post
+        assert len(env_data["test"]) == 5
+        assert isinstance(env_data["installed_packages"], list)
+        pyinfo = env_data["python"]
+        assert isinstance(pyinfo["version_info"], list)
+        assert pyinfo["version"]
+        assert pyinfo["executable"]
+    assert "write json report at: {}".format(json_path) == result.outlines[-1]
 
 
 def test_developz(initproj, cmd):
@@ -486,7 +512,7 @@ def test_developz(initproj, cmd):
         },
     )
     result = cmd("-vv", "--develop")
-    assert not result.ret
+    result.assert_success()
     assert "sdist-make" not in result.out
 
 
@@ -501,7 +527,7 @@ def test_usedevelop(initproj, cmd):
         },
     )
     result = cmd("-vv")
-    assert not result.ret
+    result.assert_success()
     assert "sdist-make" not in result.out
 
 
@@ -520,18 +546,18 @@ def test_usedevelop_mixed(initproj, cmd):
 
     # running only 'dev' should not do sdist
     result = cmd("-vv", "-e", "dev")
-    assert not result.ret
+    result.assert_success()
     assert "sdist-make" not in result.out
 
     # running all envs should do sdist
     result = cmd("-vv")
-    assert not result.ret
+    result.assert_success()
     assert "sdist-make" in result.out
 
 
 @pytest.mark.parametrize("skipsdist", [False, True])
 @pytest.mark.parametrize("src_root", [".", "src"])
-def test_test_usedevelop(cmd, initproj, src_root, skipsdist, monkeypatch):
+def test_test_usedevelop(cmd, initproj, src_root, skipsdist):
     name = "example123-spameggs"
     base = initproj(
         (name, "0.5"),
@@ -558,13 +584,13 @@ def test_test_usedevelop(cmd, initproj, src_root, skipsdist, monkeypatch):
         },
     )
     result = cmd("-v")
-    assert not result.ret
+    result.assert_success()
     assert re.match(
         r".*generated\W+xml\W+file.*junit-python\.xml" r".*\W+1\W+passed.*", result.out, re.DOTALL
     )
     assert "sdist-make" not in result.out
     result = cmd("-epython")
-    assert not result.ret
+    result.assert_success()
     assert "develop-inst-noop" in result.out
     assert re.match(
         r".*\W+1\W+passed.*" r"summary.*" r"python:\W+commands\W+succeeded.*",
@@ -573,23 +599,22 @@ def test_test_usedevelop(cmd, initproj, src_root, skipsdist, monkeypatch):
     )
 
     # see that things work with a different CWD
-    monkeypatch.chdir(base.dirname)
-    result = cmd("-c", "{}/tox.ini".format(name))
-    assert not result.ret
-    assert "develop-inst-noop" in result.out
-    assert re.match(
-        r".*\W+1\W+passed.*" r"summary.*" r"python:\W+commands\W+succeeded.*",
-        result.out,
-        re.DOTALL,
-    )
-    monkeypatch.chdir(base)
+    with base.dirpath().as_cwd():
+        result = cmd("-c", "{}/tox.ini".format(name))
+        result.assert_success()
+        assert "develop-inst-noop" in result.out
+        assert re.match(
+            r".*\W+1\W+passed.*" r"summary.*" r"python:\W+commands\W+succeeded.*",
+            result.out,
+            re.DOTALL,
+        )
 
     # see that tests can also fail and retcode is correct
     testfile = py.path.local("tests").join("test_hello.py")
     assert testfile.check()
     testfile.write("def test_fail(): assert 0")
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert "develop-inst-noop" in result.out
     assert re.match(
         r".*\W+1\W+failed.*" r"summary.*" r"python:\W+commands\W+failed.*", result.out, re.DOTALL
@@ -599,7 +624,7 @@ def test_test_usedevelop(cmd, initproj, src_root, skipsdist, monkeypatch):
     setup_py = py.path.local("setup.py")
     setup_py.write(setup_py.read() + " ")
     result = cmd()
-    assert result.ret, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert "develop-inst-nodeps" in result.out
 
 
@@ -621,7 +646,7 @@ def test_warning_emitted(cmd, initproj):
     """,
         },
     )
-    result = cmd()
+    cmd()
     result = cmd()
     assert "develop-inst-noop" in result.out
     assert "I am a warning" in result.err
@@ -660,7 +685,7 @@ def test_alwayscopy(initproj, cmd):
         },
     )
     result = cmd("-vv")
-    assert not result.ret
+    result.assert_success()
     assert "virtualenv --always-copy" in result.out
 
 
@@ -675,11 +700,11 @@ def test_alwayscopy_default(initproj, cmd):
         },
     )
     result = cmd("-vv")
-    assert not result.ret
+    result.assert_success()
     assert "virtualenv --always-copy" not in result.out
 
 
-@pytest.mark.skipif("sys.platform == 'win32'")
+@pytest.mark.skipif("sys.platform == 'win32'", reason="no echo on Windows")
 def test_empty_activity_ignored(initproj, cmd):
     initproj(
         "example123",
@@ -692,11 +717,11 @@ def test_empty_activity_ignored(initproj, cmd):
         },
     )
     result = cmd()
-    assert not result.ret
+    result.assert_success()
     assert "installed:" not in result.out
 
 
-@pytest.mark.skipif("sys.platform == 'win32'")
+@pytest.mark.skipif("sys.platform == 'win32'", reason="no echo on Windows")
 def test_empty_activity_shown_verbose(initproj, cmd):
     initproj(
         "example123",
@@ -705,11 +730,12 @@ def test_empty_activity_shown_verbose(initproj, cmd):
             [testenv]
             list_dependencies_command=echo
             commands={envpython} --version
+            whitelist_externals = echo
     """
         },
     )
     result = cmd("-v")
-    assert not result.ret
+    result.assert_success()
     assert "installed:" in result.out
 
 
@@ -725,7 +751,7 @@ def test_test_piphelp(initproj, cmd):
         },
     )
     result = cmd("-vv")
-    assert not result.ret, "{}\n{}".format(result.err, result.err)
+    result.assert_success()
 
 
 def test_notest(initproj, cmd):
@@ -740,10 +766,10 @@ def test_notest(initproj, cmd):
         },
     )
     result = cmd("-v", "--notest")
-    assert not result.ret
+    result.assert_success()
     assert re.match(r".*summary.*" r"py26\W+skipped\W+tests.*", result.out, re.DOTALL)
     result = cmd("-v", "--notest", "-epy26")
-    assert not result.ret
+    result.assert_success()
     assert re.match(r".*py26\W+reusing.*", result.out, re.DOTALL)
 
 
@@ -759,7 +785,7 @@ def test_notest_setup_py_error(initproj, cmd):
         },
     )
     result = cmd("--notest")
-    assert result.ret
+    result.assert_fail()
     assert re.search("ERROR:.*InvocationError", result.out)
 
 
@@ -767,7 +793,7 @@ def test_PYC(initproj, cmd, monkeypatch):
     initproj("example123", filedefs={"tox.ini": ""})
     monkeypatch.setenv("PYTHONDOWNWRITEBYTECODE", "1")
     result = cmd("-v", "--notest")
-    assert not result.ret
+    result.assert_success()
     assert "create" in result.out
 
 
@@ -775,7 +801,26 @@ def test_env_VIRTUALENV_PYTHON(initproj, cmd, monkeypatch):
     initproj("example123", filedefs={"tox.ini": ""})
     monkeypatch.setenv("VIRTUALENV_PYTHON", "/FOO")
     result = cmd("-v", "--notest")
-    assert not result.ret, result.outlines
+    result.assert_success()
+    assert "create" in result.out
+
+
+def test_setup_prints_non_ascii(initproj, cmd):
+    initproj(
+        "example123",
+        filedefs={
+            "setup.py": """\
+import sys
+getattr(sys.stdout, 'buffer', sys.stdout).write(b'\\xe2\\x98\\x83\\n')
+
+import setuptools
+setuptools.setup(name='example123')
+""",
+            "tox.ini": "",
+        },
+    )
+    result = cmd("--notest")
+    result.assert_success()
     assert "create" in result.out
 
 
@@ -791,7 +836,7 @@ def test_envsitepackagesdir(cmd, initproj):
         },
     )
     result = cmd()
-    assert result.ret == 0, "{}\n{}".format(result.err, result.out)
+    result.assert_success()
     assert re.match(r".*\nX:.*tox.*site-packages.*", result.out, re.DOTALL)
 
 
@@ -808,7 +853,7 @@ def test_envsitepackagesdir_skip_missing_issue280(cmd, initproj):
         },
     )
     result = cmd("--skip-missing-interpreters")
-    assert result.ret == 0, "{}\n{}".format(result.err, result.out)
+    result.assert_success()
     assert re.match(r".*SKIPPED:.*qwelkj.*", result.out, re.DOTALL)
 
 
@@ -823,7 +868,7 @@ def test_verbosity(cmd, initproj, verbosity):
         },
     )
     result = cmd(verbosity)
-    assert result.ret == 0, "{}\n{}".format(result.err, result.out)
+    result.assert_success()
 
     needle = "Successfully installed pkgX-0.0.5"
     if verbosity == "-vv":
@@ -832,25 +877,6 @@ def test_verbosity(cmd, initproj, verbosity):
         assert all(needle not in line for line in result.outlines), result.outlines
 
 
-def verify_json_report_format(data, testenvs=True):
-    assert data["reportversion"] == "1"
-    assert data["toxversion"] == tox.__version__
-    if testenvs:
-        for envname, envdata in data["testenvs"].items():
-            for commandtype in ("setup", "test"):
-                if commandtype not in envdata:
-                    continue
-                for command in envdata[commandtype]:
-                    assert command["output"]
-                    assert command["retcode"]
-            if envname != "GLOB":
-                assert isinstance(envdata["installed_packages"], list)
-                pyinfo = envdata["python"]
-                assert isinstance(pyinfo["version_info"], list)
-                assert pyinfo["version"]
-                assert pyinfo["executable"]
-
-
 def test_envtmpdir(initproj, cmd):
     initproj(
         "foo",
@@ -875,16 +901,16 @@ def test_envtmpdir(initproj, cmd):
     )
 
     result = cmd()
-    assert not result.ret
+    result.assert_success()
 
     result = cmd()
-    assert not result.ret
+    result.assert_success()
 
 
 def test_missing_env_fails(initproj, cmd):
     initproj("foo", filedefs={"tox.ini": "[testenv:foo]\ncommands={env:VAR}"})
     result = cmd()
-    assert result.ret == 1, "{}\n{}".format(result.err, result.out)
+    result.assert_fail()
     assert result.out.endswith(
         "foo: unresolvable substitution(s): 'VAR'."
         " Environment variables are missing or defined recursively.\n"
@@ -937,7 +963,7 @@ def test_exit_code(initproj, cmd, exit_code, mocker):
         assert call_error_name == "InvocationError"
         # quotes are removed in result.out
         # do not include "python" as it is changed to python.EXE by appveyor
-        expected_command_arg = " -c import sys; sys.exit({:d})".format(exit_code)
+        expected_command_arg = " -c 'import sys; sys.exit({:d})'".format(exit_code)
         assert expected_command_arg in call_command
         assert call_exit_code == exit_code
     else:
